####################################################################################
#
# File name:		Cal_Matrix
# Version:			0.1
# Software:		    
# Authors:			Zhifei Wang, Jiang Xu
# Website:			http://www.ece.ust.hk/~eexu/
# The copyright information of this program can be found in the file COPYRIGHT.
#
####################################################################################
import numpy as np
import math

def Calp(lam, L, alpha0, delta_Ne = 0, delta_Nh = 0):
    ''' Calculate the p value, where p=exp(-alpha*L/2)*exp(-1j*beta*L/2) '''
    neff = 4.2603 - 1.3096 * lam + 0.011578 * np.power(lam,2)
    delta_alpha = (8.5e-18 * delta_Ne  + 6.0e-18 * delta_Nh) * 0.5 * 1e-4     # Unit: 1/um!!!
    delta_n = -(8.8e-22 * delta_Ne + 8.5e-18 * math.pow(delta_Nh, 0.8))
    alpha = alpha0 + delta_alpha
    beta = 2 * (neff+delta_n) * np.pi / lam
    p = np.exp(-alpha*L/2)*np.exp(-1j*beta*L/2)
    return p

def ConvertChainToTransfer(M, NumMR = 1):
    """ Convert Chain Matrix to Transfer Matrix """
    if NumMR % 2 == 1:
        T00 = -M[1, 1] / M[1, 0]
        T01 = 1 / M[1, 0]
        T10 = (M[0, 1] * M[1, 0] - M[0, 0] * M[1, 1]) / M[1, 0]
        T11 = M[0, 0] / M[1, 0]
        T = np.matrix([[T00, T01], [T10, T11]])
    else:
        T00 = -M[0, 1] / M[0, 0]
        T01 = 1 / M[0, 0]
        T10 = (M[0, 0] * M[1, 1] - M[0, 1] * M[1, 0]) / M[0, 0]
        T11 = M[1, 0] / M[0, 0]
        T = np.matrix([[T00, T01], [T10, T11]])
    return T

def RewriteChainMatrix(M, NumMR = 1):
    """ Rewrite chain Matrix, let left-hand ports as inputs and right-hand ports as outputs """
    T00 = -M[0, 1] / M[0, 0]
    T01 = 1 / M[0, 0]
    T10 = (M[0, 0] * M[1, 1] - M[0, 1] * M[1, 0]) / M[0, 0]
    T11 = M[1, 0] / M[0, 0]
    T = np.matrix([[T00, T01], [T10, T11]])
    return T

def CalTransferMatrix(kappa, t, p, NumMR = 1, IsTransfer = True):
    """ Calculate Chain Matrix, return transfer matrix """
    M = np.matrix([[1.+0j, 0j], [0j, 1.+0j]])
    for iter in range(0, NumMR + 1):
        if iter % 2 == 1:
            H_iter = np.matrix([[(np.power(kappa[iter], 2) - np.power(t[iter], 2)) / kappa[iter], t[iter] / kappa[iter]],
                [-t[iter] / kappa[iter], 1 / kappa[iter]]])
            # if iter != 0:
            #     G_iter = np.matrix([[0, p[iter]], [1 / p[iter], 0]])
            # else:
            #     G_iter = np.identity(2)
            G_iter = np.matrix([[0, p[iter]], [1 / p[iter], 0]])
        if iter % 2 == 0:
            H_iter = np.matrix([[1 / kappa[iter], -t[iter] / kappa[iter]], [t[iter] / kappa[iter],
                (np.power(kappa[iter], 2) - np.power(t[iter], 2)) / kappa[iter]]])
            # if iter != 0:
            #     G_iter = np.matrix([[0, 1 / p[iter]], [p[iter], 0]])
            # else:
            #     G_iter = np.identity(2)
            G_iter = np.matrix([[0, 1 / p[iter]], [p[iter], 0]])
        if iter == 0:
            M = H_iter * M
        else:
            M = H_iter * G_iter * M
    if IsTransfer:
        result = ConvertChainToTransfer(M, NumMR)
        return result
    else:
        result = RewriteChainMatrix(M, NumMR)
        return result
    # result = ConvertChainToTransfer(M, NumMR)
    # return result

def CalParallelTransferMatrix(lam, kappa, t, alpha0, delta_Ne, delta_Nh,
                              L0, delta_perimeter, MR_spacing, NumMR_s = 1, NumMR_p = 1):
    """ Calculate Chain Matrix for parallel-coupled MRs, return transfer matrix or chain matrix """
    ''' calculate chain matrix M of MR group '''
    M = np.matrix([[1 + 0j, 0j], [0j, 1 + 0j]])
    PropMatrix = GenPropMatrix(lam, MR_spacing, alpha0)
    for i in range(0, NumMR_p):
        L = L0 + delta_perimeter * i
        p = np.append(np.ones(1), Calp(lam, L, alpha0, delta_Ne, delta_Nh))
        # calculate the chain matrix of the individual switching block
        IndividualMatrix = CalTransferMatrix(kappa, t, p, NumMR_s, False)
        M = (PropMatrix * IndividualMatrix) * M
    M = np.linalg.inv(PropMatrix) * M
    result = ConvertChainMatrixMRGroup(M, NumMR_s, NumMR_p)
    return result


def ConvertChainMatrixMRGroup(M, NumMR_s = 1, NumMR_p = 1):
    """ Convert Chain Matrix to Transfer Matrix """
    if NumMR_s % 2 == 1:
        T00 = M[0, 0] - M[0, 1] * M[1, 0] / M[1, 1]
        T01 = M[0, 1] / M[1, 1]
        T10 = -M[1, 0]/M[1, 1]
        T11 = 1 / M[1, 1]
        T = np.matrix([[T00, T01], [T10, T11]])
        return T
    else:
        return M

def GenPropMatrix(lam, L, alpha0, delta_Ne = 0, delta_Nh = 0):
    # p0 = Calp(lam, L, alpha0, delta_Ne, delta_Nh)   # for debugging
    p0 = Calp(lam, L, 0, 0, 0)
    P0 = np.matrix([[p0, 0],[0, np.conj(p0)]])
    return P0


def CalOneMR(kappa, t, p):
    T_drop = p*np.power(kappa,2)/(1-np.power(p*t,2))
    T_through = t + t*np.power(kappa*p,2)/(1-np.power(p*t,2))
    return (T_through, T_drop)
