####################################################################################
#
# File name:		SpectrumAnalysis
# Version:			0.1
# Software:		    
# Authors:			Zhifei Wang, Jiang Xu
# Website:			http://www.ece.ust.hk/~eexu/
# The copyright information of this program can be found in the file COPYRIGHT.
#
####################################################################################
import numpy as np
import time
import os

def FindIndex(P, P0, index0, right_flag = 1, Top_to_Down = 1):
    ''' Find the index (x-value) of a specific input power (y-value) '''
    num = len(P)
    result = 0
    if right_flag == 1:
        error = abs(P0-P[int(index0)])
        for index in range(int(index0)+1, num):
            if P[index]>P[index-1] and Top_to_Down:
                break
            if P[index] < P[index - 1] and not (Top_to_Down):
                break
            if abs(P0-P[index])<error:
                result = index
                error = abs(P0-P[index])
        # print('Cannot right find P0 = {0} dB'.format(P0))
        # raise SystemExit
    if right_flag == 0:
        error = abs(P0 - P[int(index0)])
        for index in range(int(index0)-1, -1, -1):
            if P[index] > P[index + 1] and Top_to_Down:
                break
            if P[index] < P[index + 1] and not (Top_to_Down):
                break
            if abs(P0 - P[index]) < error:
                result = index
                error = abs(P0 - P[index])
        # print('Cannot left find P0 = {0} dB'.format(P0))
        # raise SystemExit
    return result

def FindSubpeaks(P, error, threshhold):
    ''' Find all subpeaks, that is all local maximal points '''
    Num = len(P)
    subpeaks = np.array([])
    index = np.array([], dtype=int)
    for iter_lam in range(1, Num - 1):
        if (P[iter_lam] - P[iter_lam + 1] > error and
                        P[iter_lam] - P[iter_lam - 1] > error and P[iter_lam] > threshhold):
            subpeaks = np.append(subpeaks, P[iter_lam])
            index = np.append(index, iter_lam)
    return (subpeaks, index)

def DiveSubpeaks(subpeaks, subpeaks_index, delta_lam, SpacingEstimation):
    ''' Divide all subpeaks to peaks
    If two subpeaks are far away from each other, they belong to two peaks '''
    temp = 0
    N_subpeaks = len(subpeaks)
    peaks = []
    peaks_index = []
    for iter in range(1, N_subpeaks):
        if (subpeaks_index[iter] - subpeaks_index[iter - 1])*delta_lam > SpacingEstimation:
            peaks.append(subpeaks[temp:iter])
            peaks_index.append(subpeaks_index[temp:iter])
            temp = iter
    peaks.append(subpeaks[temp:N_subpeaks + 1])
    peaks_index.append(subpeaks_index[temp:N_subpeaks + 1])
    return (peaks, peaks_index)

def SpectrumAnalysis(lam, P_through, P_drop, P_through_on, P_drop_on, SpacingEstimation = 0.5, delta_dB = 10,
                     FileName = '', FindPeakError = 0, PeakThreshold = -10):
    '''
    Note input should be in the unit of dB
    Note lam should be in the unit of nm
    SpacingEstimation: estimated linewidth range
    delta_dB: the threshold of peak-valley difference
    PeakThreshold: Peak amplitude threshold
    FindPeakError: Ripple threshold between two points
    '''
    Num = len(lam)
    delta_lam = (lam[-1]-lam[0])/(Num-1)
    (Drop_on_subpeaks, Drop_on_subpeaks_index) = \
        FindSubpeaks(P_drop_on, FindPeakError, PeakThreshold)
    (Drop_subpeaks, Drop_subpeaks_index) = \
        FindSubpeaks(P_drop, FindPeakError, PeakThreshold)
    (Drop_on_peaks, Drop_on_peaks_index) = \
        DiveSubpeaks(Drop_on_subpeaks, Drop_on_subpeaks_index, delta_lam, SpacingEstimation)
    (Drop_peaks, Drop_peaks_index) = \
        DiveSubpeaks(Drop_subpeaks, Drop_subpeaks_index, delta_lam, SpacingEstimation)

    print(len(Drop_on_subpeaks), len(Drop_on_peaks))

    if len(Drop_peaks) != len(Drop_on_peaks):
        print('\n!!!! Warning !!!!\n\
            Please choose an appropriate wavelength range where the number of \n\
            on-resonant states is EQUAL to that of off-resonant state!\n\
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
        # raise SystemExit
        return (0, 0, 0, 0)

    ''' Calculate spectrum characteristics '''
    N_peaks = len(Drop_on_peaks)
    ResonantWavelength = np.array([])
    ResonantWavelength_off = np.array([])
    Wavelength_shift = np.array([])
    DropLoss = np.array([])
    ThroughLoss = np.array([])
    Crosstalk_drop = np.array([])
    Crosstalk_through = np.array([])
    BW = np.array([])
    Roll_off = np.array([])
    for iter in range(0, N_peaks):
        ''' Resonant wavelength, wavelength shift '''
        ''' If only one subpeak, choose it; else use average of max subpeaks and min subpeaks '''
        if len(Drop_on_peaks_index[iter]) == 0:
            print("The on-resonant Drop Loss is too small or even does not exist!!!")
            return (0, 0, 0, 0)
        if len(Drop_on_peaks_index[iter]) == 1:
            ResonantWavelength = np.append(ResonantWavelength, lam[Drop_on_peaks_index[iter]])
        else:
            temp = (lam[Drop_on_peaks_index[iter][-1]] + lam[Drop_on_peaks_index[iter][0]]) / 2
            ResonantWavelength = np.append(ResonantWavelength, temp)

        if len(Drop_peaks_index[iter]) == 0:
            print("The off-resonant Drop Loss is too small or even does not exist!!!")
            return (0, 0, 0, 0)
        if len(Drop_peaks_index[iter]) == 1:
            ResonantWavelength_off = np.append(ResonantWavelength_off, lam[Drop_peaks_index[iter]])
        else:
            temp = (lam[Drop_peaks_index[iter][-1]] + lam[Drop_peaks_index[iter][0]]) / 2
            ResonantWavelength_off = np.append(ResonantWavelength_off, temp)
        Wavelength_shift = np.append(Wavelength_shift, ResonantWavelength[-1] - ResonantWavelength_off[-1])
        ''' Calculate Loss '''
        ''' If only one subpeak, consider this point only; else use average between max subpeaks and min subpeaks '''
        if len(Drop_on_peaks_index[iter]) == 1:
            temp_AveDropLoss = P_drop_on[Drop_on_peaks_index[iter][0]]
            DropLoss = np.append(DropLoss, temp_AveDropLoss)
            ThroughLoss = np.append(ThroughLoss, P_through[Drop_on_peaks_index[iter][0]])
        else:
            temp_AveDropLoss = np.average(P_drop_on[Drop_on_peaks_index[iter][0]:Drop_on_peaks_index[iter][-1]+1])
            DropLoss = np.append(DropLoss, temp_AveDropLoss)
            temp = np.average(P_through[Drop_on_peaks_index[iter][0]:Drop_on_peaks_index[iter][-1]+1])
            ThroughLoss = np.append(ThroughLoss, temp)
        ''' Calculate 3-dB BW '''
        temp1 = FindIndex(P_drop_on, temp_AveDropLoss - 3, Drop_on_peaks_index[iter][-1], right_flag=1)
        temp2 = FindIndex(P_drop_on, temp_AveDropLoss - 3, Drop_on_peaks_index[iter][0], right_flag=0)
        if temp1 != 0 and temp2 != 0:
            temp = (temp1 - temp2) * delta_lam
            BW = np.append(BW, temp)
            ''' Calculate crosstalk '''
            Crosstalk_drop_iter = np.average(P_drop[temp2:temp1+1])
            Crosstalk_through_iter = np.average(P_through_on[temp2:temp1+1])
            Crosstalk_drop = np.append(Crosstalk_drop, Crosstalk_drop_iter)
            Crosstalk_through = np.append(Crosstalk_through, Crosstalk_through_iter)
            ''' Calculate Roll-off Speed (Make sure find -3dB first!)'''
            loc_10dB_right = FindIndex(P_drop_on, Drop_on_peaks[iter][-1] - delta_dB, Drop_on_peaks_index[iter][-1],
                              right_flag=1)
            loc_10dB_left = FindIndex(P_drop_on, Drop_on_peaks[iter][0] - delta_dB, Drop_on_peaks_index[iter][0], right_flag=0)
            loc_1dB_right = FindIndex(P_drop_on, Drop_on_peaks[iter][-1] - 1, Drop_on_peaks_index[iter][-1],
                              right_flag=1)
            loc_1dB_left = FindIndex(P_drop_on, Drop_on_peaks[iter][0] - 1, Drop_on_peaks_index[iter][0], right_flag=0)
            if loc_10dB_right != 0 and loc_10dB_left != 0 and loc_1dB_right and loc_1dB_left:
                Roll_off_right = (delta_dB+1)/((loc_10dB_right-loc_1dB_right)*delta_lam) # dB/nm
                Roll_off_left = (delta_dB+1) / ((loc_1dB_left-loc_10dB_left)*delta_lam) # dB/nm
                temp = (Roll_off_right + Roll_off_left) / 2
                Roll_off = np.append(Roll_off, temp)
            else:
                print('No approxiate data to calculate Roll-off speed!!')

    AveWavelength_shift = np.average(Wavelength_shift)
    MinDropLoss = min(DropLoss)
    MinThroughLoss = min(ThroughLoss)
    AveDropLoss = np.average(DropLoss)
    AveThroughLoss = np.average(ThroughLoss)
    AveCrosstalk_drop = np.average(Crosstalk_drop)
    AveCrosstalk_through = np.average(Crosstalk_through)
    AveBW = np.average(BW)
    AveRoll_off = np.average(Roll_off)

    print("\n---------------- Spectrum Analysis Result ------------------")
    print("Number of Resonant wavelength = ", len(ResonantWavelength))
    write_string = ''
    for iter in ResonantWavelength:
        write_string += ("{0:.3f}nm, ".format(iter))
    write_string = write_string[0:-2]
    print("Resonant wavelength = {0}".format(write_string))
    print("Average wavelength shift between on and off state = {0:.3f} (nm) ".format(AveWavelength_shift))
    # print("Minimal Drop Loss = {0:.3f} dB".format(MinDropLoss))
    print("Average Drop Loss = {0:.3f} dB".format(AveDropLoss))
    print("Average Through Loss = {0:.3f} dB".format(AveThroughLoss))
    print("Average Crosstalk at Drop port = {0:.3f} dB".format(AveCrosstalk_drop))
    print("Average Crosstalk at Through port = {0:.3f} dB".format(AveCrosstalk_through))
    print("Average 3-dB Bandwidth = {0:.3f} nm".format(AveBW))
    print("Average Roll-off Speed = {0:.3f} dB/nm".format(AveRoll_off))
    print("---------------------- End of Result -----------------------\n")

    ''' Write File '''
    if FileName != '':
        PathName = 'ResultData/'
        if not os.path.exists(PathName):
            os.makedirs(PathName)
        f = open(PathName + FileName+'_SpectrumAnalysis.txt', 'w')
        f.write("---------------- Spectrum Analysis Result ------------------\n")
        f.write("** By Zhifei , "+time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())+" **\n")
        f.write("------------------------------------------------------------\n")
        # f.write("---------------- Spectrum Analysis Result ------------------\n")
        f.write("Number of Resonant wavelength = {0:d}\n".format(len(ResonantWavelength)))
        f.write("Average wavelength shift between on and off state = {0:.3f} (nm)\n".format(AveWavelength_shift))
        # f.write("Minimal Drop Loss = {0:.3f} dB\n".format(MinDropLoss))
        f.write("Average Drop Loss = {0:.3f} dB\n".format(AveDropLoss))
        f.write("Average Through Loss = {0:.3f} dB\n".format(AveThroughLoss))
        f.write("Average Crosstalk at Drop port = {0:.3f} dB\n".format(AveCrosstalk_drop))
        f.write("Average Crosstalk at Through port = {0:.3f} dB\n".format(AveCrosstalk_through))
        f.write("Average 3-dB Bandwidth = {0:.3f} nm\n".format(AveBW))
        f.write("Average Roll-off Speed = {0:.3f} dB/nm\n".format(AveRoll_off))

        f.write("\n----------------------- Detailed Result --------------------\n")
        f.write("Resonant wavelength: \n ")
        write_string = ''
        for iter in ResonantWavelength:
            write_string += ("{0:.3f}nm, ".format(iter))
        write_string = write_string[0:-2] + '\n\n'
        f.write(write_string)

        f.write("Wavelength shift:\n ")
        write_string = ''
        for iter in Wavelength_shift:
            write_string += ("{0:.3f}nm, ".format(iter))
        write_string = write_string[0:-2] + '\n'
        f.write(write_string)
        f.write("---------------------- End of Result -----------------------\n\n")

    return (AveDropLoss, AveThroughLoss, AveCrosstalk_drop, AveCrosstalk_through)


